from math import log, sqrt
import os
import pandas as pd
import aligngraph
from aligngraph import convert_mismatches, AlnGraph
import networkx as nx
from Bio import SeqIO


edge_l = []
node_l = []
node_id_dict = {}

def get_nd_no(nd_id):
    """
    map the node id back to a numerical id if the key exists
    """
    try:
        return node_id_dict[nd_id]
    except:
        print("The id of node not in the dictionary")
        return -1
        
def get_nd_id(nd_no):
    try:
        return [id for id, no in node_id_dict.items() if no == nd_no]
    except:
        print("The no of node not in the dictionary")

def write_lines(lines, out_file):
    with open(out_file, 'w') as out_f:
        for line in lines:
            print(line, file = out_f)    
                
def nx2gfa(nx_graph, out_f=None, folder=None):
    os.makedirs(folder, exist_ok=True)
    lines = []
    n_nd = nx_graph.number_of_nodes()
    n_ed = nx_graph.number_of_edges()
    lines.append("V:\t" + str(n_nd) + "\t" + "E:\t" + str(n_ed))
    for nd, attr in nx_graph.nodes(data=True):
        line = "V\t" + str(nd) + '\t' + attr['base']
        lines.append(line)
    for u, v, attr in nx_graph.edges(data=True):
        line = "E\t" + str(u) + '\t' + str(v) + '\t:' + str(attr['weight'])
        lines.append(line)
    
    folder_path = folder if folder else ""
    out_file = os.path.join(folder_path, "aln_graph.gfa") if out_f is None else os.path.join(folder_path, out_f)
    write_lines(lines, out_file)

def aln2nx(aln_graph): 
    global edge_l, node_l, node_id_dict
    node_l = list(aln_graph.nodes.values())
    edge_l = list(aln_graph.edges.values())

    nx_g = nx.DiGraph()
    #dict stores the id and the number id of each node
    node_id_dict = {}
    for i, nd in enumerate(node_l):
        node_id_dict[nd.ID] = i
        nx_g.add_node(i, base = nd.base)
        if nd.base == 'B':
            src = i
        elif nd.base == 'E':
            des = i      
    for i, edge in enumerate(edge_l):
        #src = node_id_dict[edge.in_node.ID]
        #des = node_id_dict[edge.out_node.ID]
        src = get_nd_no(edge.in_node.ID)
        des = get_nd_no(edge.out_node.ID)
        if src == -1 or des == -1:
            print(i," missing")
        nx_g.add_edge(src, des, weight = edge.count)
    return [nx_g, src, des]


def nx_path_score(nx_graph, path):
    # input: a nx graph with edge property 'weight'; a path
    # return: sum weight of this path
    def path2iter(path):
        for i in range(len(path)-1):
            yield (path[i], path[i+1])
    ed_l = path2iter(path)
    score = sum(nx_graph.get_edge_data(*ed)['weight'] for ed in ed_l )
    return score

def nx_homo(nx_g, out_f):
    #count the homopolymer region and return a list of range
    homo_loc = []
    min_len = 3
    win_len = 10
    topo_list = list(nx.topological_sort(nx_g)) 
    #print(topo_list[0:30])
    for i, nd in enumerate(topo_list[10:]):
        #nd_l = [u for u,v, attr  in nx_g.in_edges(nd, data = True)]
        nd_l = []
        for u,v in nx_g.in_edges(nd):
            nd_l.append(u)
        in_base_l = [nx_g.nodes[nd]['base'].upper() for nd in nd_l]
        char_l = ['A', 'C', 'G', 'T']
        freq_l = []
        freq_l.append(in_base_l.count('A'))
        freq_l.append(in_base_l.count('C'))
        freq_l.append(in_base_l.count('G'))
        freq_l.append(in_base_l.count('T'))
        
        if max(freq_l) >= min_len: #it's a homopolymer region of length min_len
            #print(i, nd, nd_l, freq_l)
            homo_end = nd
            homo_begin = 0
            homo_char = char_l[freq_l.index(max(freq_l))] #base of this hp region
            #now locate the beginning of this region
            #win = topo_list[ i - win_len + 1: i + 1]
            win = topo_list[ 10 + i - win_len + 1: 10 + i] #i is the index start from 10 in topo-list
            
            for win_nd in win[::-1]:
                if nx_g.nodes[win_nd]['base'].upper() != homo_char:
                    homo_begin = win_nd
                    homo_len = topo_list.index(homo_end) - topo_list.index(homo_begin) - 1 #exclude the first base that's not in homopolymer region
                    break
            homo_loc.append((homo_begin, homo_end, homo_char, homo_len))
    #print("homopolymer regions:", len(homo_loc),"\n", homo_loc)    
    homo_df = pd.DataFrame(homo_loc, columns = ['homo_begin', 'homo_end', 'homo_char', 'homo_len'])
    
    homo_df.to_csv(out_f)

    return homo_df
    
class AlignGraphUtilError(Exception):
    pass

def phi_coeff(xvec, yvec):
    nx1 = sum( xvec == 1 )
    nx0 = len(xvec) - nx1
    ny1 = sum( yvec == 1 )
    ny0 = len(yvec) - ny1
    xyvec = (xvec << 1) + yvec
    n11 = sum( xyvec == 3 )
    n10 = sum( xyvec == 2 )
    n01 = sum( xyvec == 1 )
    n00 = sum( xyvec == 0 )
    return (n11 * n00 - n10 * n01) / sqrt( nx1 * nx0 * ny1 * nx0 + 1)

class Simple_Alignment_Hit(object):
    """ A simple class to wrap the output of the blasr "-m 5" option """
    def __init__(self, rm5_line):
        rm5_line = rm5_line.strip().split()
        self.query_id = rm5_line[0]
        self.query_length = int(rm5_line[1])
        self.query_start = int(rm5_line[2])
        self.query_end = int(rm5_line[3])
        self.query_strand = rm5_line[4]
        self.target_id = rm5_line[5]
        self.target_length = int(rm5_line[6])
        self.target_start = int(rm5_line[7])
        self.target_end = int(rm5_line[8])
        self.target_strand = rm5_line[9]
        self.alignedQuery = rm5_line[16]
        self.alignedTarget = rm5_line[18]

def simple_align_hit_iterator(rm1_fn, ref_group = None):
    with open(rm1_fn) as f:
        for l in f:
            ll = l.strip().split()
            if ref_group != None and ll[5] != ref_group:
                continue
            yield Simple_Alignment_Hit(l)

def get_aln_array(aln_iter,
                  max_num_reads = None, 
                  remove_in_del = False, 
                  min_length = None):

    rMap = dict(zip("ACGTacgtNn-", "TGCAtgcaNn-"))

    alns = []
    nread = 0
    i = 0
    bp = {}
    reads = {}

    for h in aln_iter:

        if min_length != None:
            if h.target_end - h.target_start < min_length:
                continue

        nread += 1

        if max_num_reads != None:
            if nread > max_num_reads:
                break

        read_id = h.query_id 

        if read_id in reads: continue #avoid duplicate readId in the data
        
        ts = h.target_start
        te = h.target_end
        qs = h.query_start
        qe = h.query_end

        alnT, alnQ = h.alignedTarget.upper(), h.alignedQuery.upper()
        dqPos = 1
        
        if h.target_strand == '-':
            alnT = "".join([rMap[c] for c in alnT[::-1]])
            alnQ = "".join([rMap[c] for c in alnQ[::-1]])
            dqPos = -1
            qs, qe = qe - 1, qs - 1

        if remove_in_del:
            aln_pair = zip(alnQ, alnT)
            new_aln_pair = []
            for b1, b2 in aln_pair:
                if b1 == "-":
                    b1 = b2
                if b2 != "-":
                    new_aln_pair.append( (b1,b2) )
            alnQ, alnT = zip(*new_aln_pair)
            alnQ = "".join(alnQ)
            alnT = "".join(alnT)

        alnQ, alnT = convert_mismatches(alnQ,alnT)

        if alnQ[0] == "-" or alnT[0] == "-":
            continue
        if alnQ[-1] == "-" or alnT[-1] == "-":
            continue

        #print h.target_strand 
        #print "r %10d %10d" % (qs, qe), alnQ
        #print "t %10d %10d" % (ts, te), alnT
        #print

        bp[read_id] = ( (qs, qe), ts)
        alns.append(  ( ( qs, qe, alnQ ), ( ts, te, alnT ) , read_id ) )
        tPos = ts
        qPos = qs
        assert len(alnQ) == len(alnT)
        reads[read_id] = alnQ.replace("-","")
        i += 1

    return alns

def construct_aln_graph_from_fasta(read_fasta_fn, 
                                    backbone_fasta_fn, 
                                    max_num_reads = None, 
                                    remove_in_del = False, 
                                    ref_group = None,
                                    min_length = None):

    import os

    os.system("blasr %s %s -m 5 --nproc 16 --out %s" % (read_fasta_fn, backbone_fasta_fn, read_fasta_fn+".aln_unsorted"))
    os.system("cat %s | sort > %s" % (read_fasta_fn+".aln_unsorted", read_fasta_fn+".aln" ))
    
    aln_hit_iterator = simple_align_hit_iterator(read_fasta_fn+".aln", ref_group = ref_group)

    alns = get_aln_array(aln_hit_iterator, 
                         max_num_reads = max_num_reads, 
                         remove_in_del = remove_in_del, 
                         min_length = min_length)
        
    backboneSeq = open(backbone_fasta_fn).read()
    backboneSeq = "".join(backboneSeq.split('\n')[1:])

    g = AlnGraph(backboneSeq)

    i = 0
    for aln in alns:
        rId = aln[2]
        rId = rId.split("/")[0]
        aln = aln[0:2]
        g.add_alignment( aln, "%s" % rId)

    return g

def sorted_nodes(g):
    return g.get_sorted_nodes()

def read_node_vector(g, entropy_th = 2.5):
    read_to_nodes, high_entropy_nodes = g.get_read_node_vector(entropy_th=entropy_th)
    return read_to_nodes, high_entropy_nodes

def clustering_read( read_to_nodes, high_entropy_nodes, k_cluster = 2, random_seed = 42, cleanup_th = 0.5):
    
    import random
    random.seed(random_seed)

    cluster = {}
    read_to_binary_vector = {}
    count = 0
    for r in read_to_nodes:   
        read_to_binary_vector[r] = np.array([ 1 if c != "-" else -1 for c in read_to_nodes[r] ])
        k = random.randint(0, k_cluster-1)
        #k = count % k_cluster
        cluster.setdefault(k,[])
        cluster[k].append(r)
        count += 1
    
    n_iteration = 0
    cluster_vec = {}
    while 1:
        for k in range(k_cluster):
            new_vec = np.zeros(len(high_entropy_nodes),dtype=np.float)
            #print k, len(cluster[k])
            for r in cluster[k]:
                new_vec += read_to_binary_vector[r]
            new_vec /=  (len(cluster[k])+1)
            cluster_vec[k] = np.array([1 if v>=0 else -1 for v in new_vec])

        #avoid degnerated cluster    
        for k in range(k_cluster):
            for j in range(k+1, k_cluster):
                if sum(cluster_vec[j] == cluster_vec[k]) > 0.5*len(high_entropy_nodes):
                    cluster_vec[j] = np.array([random.choice([-1,1]) for v in new_vec])

        cluster = {}
        for r in read_to_nodes:   
            distances = []
            for k in range(k_cluster):
                cluster.setdefault(k,[])
                distances.append( (sum(read_to_binary_vector[r] * cluster_vec[k]), k) )

            distances.sort()

            cluster[distances[-1][1]].append(r)
        
        n_iteration += 1
        if n_iteration  > 10:
            break
            
    cluster = {}
    for r in read_to_nodes:   
        distances = []
        for k in range(k_cluster):
            cluster.setdefault(k,[])
            distances.append( (sum(read_to_binary_vector[r] * cluster_vec[k]), k) )

        distances.sort()
        if distances[-1][0] > cleanup_th*len(high_entropy_nodes):
            cluster[distances[-1][1]].append(r)
    return cluster, cluster_vec

def sorted_node_data(aln_graph, entropy_th = 0.5, interval = None):
    ne, hne = aln_graph.get_high_entropy_nodes(coverage_th=0)
    node_to_entropy = dict( [ (v[1],v[2]) for v in ne ] ) 
    read_ids = set()
    for n in aln_graph.nodes.values():
        for r in n.info:
            read_ids.add(r)
    read_id_to_pos = dict(( (x[1],x[0]) for x in enumerate(list(read_ids))) )

    backbone_node_to_pos =  aln_graph.backbone_node_to_pos 

    data = []
    for n in sorted_nodes(aln_graph):
        s = [" "] * len(read_id_to_pos)
        for r in n.info:
            s[read_id_to_pos[r]] = n.base

        if n.base not in ["B", "E"]:
            bpos = backbone_node_to_pos[n.backbone_node]
            if interval != None and (bpos < interval[0] or bpos > interval[1]):
                continue
        ent = node_to_entropy[n] if n in node_to_entropy else 0
        if n.base not in ["B","E"] and ent >= entropy_th:
            data.append ( (  backbone_node_to_pos[n.backbone_node],\
                  "+" if n in aln_graph.consensus_path else "-",\
                  "+" if n.is_backbone == True else "-",\
                  n.base, "".join(s), len(n.info),\
                  n.backbone_node.coverage,\
                  node_to_entropy[n] if n in node_to_entropy else "-" ) )
    return data

def detect_missing(aln_graph, entropy_th = 0.66, interval = None):
    data = sorted_node_data(aln_graph, entropy_th = 0, interval = interval)
    s = []
    for d in data:
        if d[7] > entropy_th and d[2] == "-":
            s.append(d[3].lower())
            continue
        if d[2] =="+" and d[7] > entropy_th:
            #pass
            s.append(d[3].lower())
        elif d[2] == "+":
            s.append(d[3].upper())
    return "".join(s)

def mark_lower_case_base(aln_graph, entropy_th = 0.66, interval = None):
    data = sorted_node_data(aln_graph, entropy_th = 0, interval = interval)
    s = []
    for d in data:
        #if d[7] > entropy_th and d[2] == "-":
        #    s.append(d[3].lower())
        #    continue
        if d[2] =="+" and d[7] > entropy_th:
            #pass
            s.append(d[3].lower())
        elif d[2] == "+":
            s.append(d[3].upper())
    return "".join(s)
    
def merge_filter(dbs_file, samp_file, filter_file):
    # merge the output of two modules and filter out short ones (mainly from DBS).
    max_len = 0
    thres = 100
    for record in SeqIO.parse(dbs_file, "fasta"):
        if len(record.seq) > max_len:
            max_len = len(record.seq)
    for record in SeqIO.parse(samp_file, "fasta"):
        if len(record.seq) > max_len:
            max_len = len(record.seq)

    #print("longest seq = ", max_len)
    filtered_seqs = []
    for record in SeqIO.parse(dbs_file, "fasta"):
        if len(record.seq) > max_len - thres:
            filtered_seqs.append(record)
    for record in SeqIO.parse(samp_file, "fasta"):
        if len(record.seq) > max_len - thres:
            filtered_seqs.append(record)
    SeqIO.write(filtered_seqs, filter_file, "fasta")

    return

def extract_sequences(filepath):
    """从 .fa 文件中提取所有序列内容"""
    sequences = []
    with open(filepath, 'r') as f:
        for line in f:
            if not line.startswith('>'):  # 跳过标题行
                sequences.append(line.strip())
    return ''.join(sequences)

def are_sequences_identical(file1, file2):
    """比较两个 .fa 文件的序列是否一致"""
    seq1 = extract_sequences(file1)
    seq2 = extract_sequences(file2)
    return seq1 == seq2


import networkx as nx
import numpy as np
# from numba import cuda
import torch
# cuda.select_device(0)
# cuda.get_current_device().reset()

# from numba import config
# config.CUDA_IR_VERSION = 1.6

def is_node_redundant(graph, node):

    for predecessor ,successor in graph.predecessors(node):
        if graph.edges[predecessor, node]['weight'] >= 5 and graph.edges[node, successor]['weight'] >= 5:
            return False
    
    return True

def max_weight_path_gpu(graph, start, end):

    nodes = list(graph.nodes)
    edges = list(graph.edges(data='weight'))
    node_indices = {node: i for i, node in enumerate(nodes)}

    node_type_mapping = {node_type: i for i, node_type in enumerate(
        {graph.nodes[node]['type'] for node in nodes}
    )}
    node_types = torch.tensor(
        [node_type_mapping[graph.nodes[node]['type']] for node in nodes],
        dtype=torch.int32
    ).to('cuda')

    dp = torch.full((len(nodes),), float('-inf'), device='cuda')  # 在 GPU 上
    dp[node_indices[start]] = 0
    path = torch.full((len(nodes),), -1, dtype=torch.int32, device='cuda')

    edge_indices = torch.tensor(
        [[node_indices[u], node_indices[v]] for u, v, _ in edges], device='cuda'
    )
    edge_weights = torch.tensor(
        [float(data) if data is not None else 0.0 for _, _, data in edges],
        dtype=torch.float32, device='cuda'
    )

    for _ in range(len(nodes)):
        for i, (u, v) in enumerate(edge_indices):
            weight = edge_weights[i]
            if dp[u] + weight > dp[v]:
                dp[v] = dp[u] + weight
                path[v] = u

    max_weight_path = []
    node_labels = []
    current_index = node_indices[end]
    while current_index != -1:
        node = nodes[current_index]
        max_weight_path.append(node)
        node_labels.append(node_types[current_index].item())
        current_index = path[current_index].item()

    max_weight_path.reverse()
    node_labels.reverse()

    max_weight_graph = nx.DiGraph()
    for i in range(len(max_weight_path)):
        node = max_weight_path[i]
        node_type = graph.nodes[node]['type']
        max_weight_graph.add_node(node, type=node_type)
        if i > 0:
            prev_node = max_weight_path[i - 1]
            weight = graph.edges[prev_node, node]['weight']
            max_weight_graph.add_edge(prev_node, node, weight=weight)

    return dp[node_indices[end]].item(), max_weight_path, node_labels, max_weight_graph
def max_weight_path(graph):
    dp = {node: float('-inf') for node in graph}
    path = {node: None for node in graph}
    node_types = {node: graph.nodes[node]['type'] for node in graph}

    start_node = next(iter(graph.nodes))
    dp[start_node] = 0

    
    for node in graph:
        
        for neighbor in graph.successors(node):
            weight = graph.edges[node, neighbor]['weight']
            
            if dp[node] + weight > dp[neighbor]:
                dp[neighbor] = dp[node] + weight
                path[neighbor] = node

    end_node = max(dp, key=dp.get)
    max_weight = dp[end_node]


    current_node = end_node
    while current_node is not None:
        if dp[current_node] == 0:
            start_node = current_node
            break
        current_node = path[current_node]

   
    max_weight_path = []
    node_labels = [[]]
    current_node = end_node

    while current_node is not None:
        max_weight_path.append(current_node)
        node_labels[0].append(node_types[current_node])
        current_node = path[current_node]
    max_weight_path.reverse()
    node_labels.reverse()

   
    max_weight_graph = nx.DiGraph()
    for node in max_weight_path:
        node_type = graph.nodes[node].get('type')
        max_weight_graph.add_node(node, type=node_type)  
        if path[node] is not None:
            weight = graph.edges[path[node], node].get('weight') 
            max_weight_graph.add_edge(path[node], node, weight=weight)

    return max_weight, max_weight_path, node_labels, max_weight_graph


def max_weight_path_old(graph, start, end):
    dp = {node: float('-inf') for node in graph}
    dp[start] = 0  # 起点权值设为0
    path = {node: None for node in graph}
    node_types = {node: graph.nodes[node]['type'] for node in graph}

    
    for node in graph:
        
        for neighbor in graph.successors(node):
            weight = graph.edges[node, neighbor]['weight']
            
            if dp[node] + weight > dp[neighbor]:
                dp[neighbor] = dp[node] + weight
                path[neighbor] = node

    
    max_weight_path = []
    node_labels = [[]]
    current_node = end
    while current_node is not None:
        max_weight_path.append(current_node)
        node_labels[0].append(node_types[current_node])
        current_node = path[current_node]
    max_weight_path.reverse()
    node_labels.reverse()


    max_weight_graph = nx.DiGraph()
    for node in max_weight_path:
        node_type = graph.nodes[node].get('type') 
        max_weight_graph.add_node(node, type=node_type) 
        if path[node] is not None:
            weight = graph.edges[path[node], node].get('weight')  
            max_weight_graph.add_edge(path[node], node, weight=weight)

    return dp[end], max_weight_path, node_labels, max_weight_graph

 
def max_ratio_paths(graph, start, end):
    dp = {node: float('-inf') for node in graph}
    dp[start] = 0   
    paths = {node: [] for node in graph}   
    paths[start] = [[start]]  

    node_types = {node: graph.nodes[node]['type'] for node in graph}  
    path_counts = {node: 0 for node in graph}   
    
    for node in graph:
        
        for neighbor in graph.successors(node):
            weight = graph.edges[node, neighbor]['weight']
             
            ratio = (dp[node] + weight) / (path_counts[node] + 1)
            
            if path_counts[neighbor] == 0 or ratio > dp[neighbor] / path_counts[neighbor]:
                dp[neighbor] = dp[node] + weight
                paths[neighbor] = [path + [neighbor] for path in paths[node]]
                path_counts[neighbor] = path_counts[node] + 1
             
            elif ratio == dp[neighbor] / path_counts[neighbor]:
                paths[neighbor].extend([path + [neighbor] for path in paths[node]])
                path_counts[neighbor] += path_counts[node]
    max_paths = paths[end]
     
    node_labels = [[node_types[node] for node in path] for path in max_paths]

    return dp[end], max_paths, node_labels, path_counts[end]





# @cuda.jit
# def update_dp_path(dp, path, weights, successors, nodes_per_level, level_offset):
#     idx = cuda.grid(1)
#     if idx >= nodes_per_level:
#         return

     
#     current_node = level_offset + idx
#     for j in range(successors[current_node][0]):  
#         neighbor = successors[current_node][j + 1]
#         weight = weights[current_node][j]
        
        
#         cuda.atomic.max(dp, neighbor, dp[current_node] + weight)
#         if dp[neighbor] == dp[current_node] + weight:
#             path[neighbor] = current_node

 
# def max_weight_path_cuda(graph):
    
#     node_to_index = {node: idx for idx, node in enumerate(graph.nodes())}
#     index_to_node = {idx: node for node, idx in node_to_index.items()}

    
#     max_degree = max(len(list(graph.neighbors(node))) for node in graph)
#     start_node = next(node for node in graph if graph.in_degree(node) == 0)
#     start_node_index = node_to_index[start_node]
 
#     n_nodes = len(graph)
#     dp = np.full(n_nodes, -np.inf, dtype=np.float32)
#     path = np.full(n_nodes, -1, dtype=np.int32)

#     weights = np.zeros((n_nodes, max_degree), dtype=np.float32)
#     successors = np.zeros((n_nodes, max_degree + 1), dtype=np.int32)

 
#     for node in graph:
#         node_idx = node_to_index[node]
#         neighbors = list(graph.neighbors(node))
#         successors[node_idx, 0] = len(neighbors)
#         for j, neighbor in enumerate(neighbors):
#             neighbor_idx = node_to_index[neighbor]
#             successors[node_idx, j + 1] = neighbor_idx
#             weights[node_idx, j] = graph.edges[node, neighbor]['weight']

 
#     dp[start_node_index] = 0

 
#     d_dp = cuda.to_device(dp)
#     d_path = cuda.to_device(path)
#     d_weights = cuda.to_device(weights)
#     d_successors = cuda.to_device(successors)
 
#     threads_per_block = 256
#     blocks_per_grid = (n_nodes + threads_per_block - 1) // threads_per_block
 
#     levels = [(0, n_nodes)]
#     for level_offset, nodes_per_level in levels:
#         update_dp_path[blocks_per_grid, threads_per_block](
#             d_dp, d_path, d_weights, d_successors, nodes_per_level, level_offset
#         )
 
#     dp = d_dp.copy_to_host()
#     path = d_path.copy_to_host()
 
#     end_node_index = np.argmax(dp)
#     max_weight_path = []
#     while end_node_index != -1:
#         max_weight_path.append(index_to_node[end_node_index])
#         end_node_index = path[end_node_index]
#     max_weight_path.reverse()

#     return max_weight_path, dp[np.argmax(dp)]


import consensus as CS4A
import argparse
def run_consensus_iteration(reads, initial_reference, num_iterations, output_folder):
    current_reference = initial_reference
    for i in range(num_iterations):
        print(f"Iteration {i+1}/{num_iterations}")
        output_file = CS4A.get_consensus_sequence(reads, current_reference, f"{output_folder}/iteration_{i+1}")
        current_reference = output_file
    return current_reference

def main():
    parser = argparse.ArgumentParser(description="Generate consensus sequence from reads")
    parser.add_argument("-b", required=True, help="Path to the initial reference file")
    parser.add_argument("-r", required=True, help="Path to the input reads file")
    parser.add_argument("-outdir", required=True, help="Path to the output folder")
    parser.add_argument("-n", type=int, default=3, help="Number of iterations to perform")
    args = parser.parse_args()

    final_consensus = run_consensus_iteration(args.r, args.b, args.n, args.outdir)
    print(f"Final consensus sequence saved at: {final_consensus}")

if __name__ == "__main__":
    main()


import networkx as nx

def is_node_redundant(graph, node):
    for predecessor in graph.predecessors(node):
        if graph.edges[predecessor, node]['weight'] > 2:
            return False
    
    for successor in graph.successors(node):
        if graph.edges[node, successor]['weight'] > 2:
            return False
    
    
    if graph.in_degree(node) == 0 and graph.out_degree(node) >= 2:
        for successor in graph.successors(node):
            if graph.edges[node, successor]['weight'] > 2:
                return False
    
   
    if graph.out_degree(node) == 0 and graph.in_degree(node) >= 2:
        for predecessor in graph.predecessors(node):
            if graph.edges[predecessor, node]['weight'] > 2:
                return False
    
    return True


def polish(graph):
    redundant_nodes = []
    node_labels = {}
    
    for node in graph.nodes:
        if is_node_redundant(graph, node):
            redundant_nodes.append(node)
        else:
            node_labels[node] = graph.nodes[node]['type']
    
    node_types = list(node_labels.values())
    node_types.reverse()
    return graph, [node_types]



import networkx as nx
def read_graph_from_file(filename):
    graph = nx.DiGraph()
    with open(filename, 'r') as file:
        lines = file.readlines()

       
        line = lines[0].strip().split('\t')
        num_vertices = int(line[1])
        num_edges = int(line[3])

       

        for line in lines[1 :1 +  num_vertices]:
            line = line.strip().split('\t')
            vertex_id = line[1]

            vertex_type = line[2]
            graph.add_node(vertex_id, type=vertex_type)

       
        for line in lines[1 + num_vertices:]:
            line = line.strip().split('\t')
            edge_type = line[0]
            start_vertex = line[1]
            end_vertex = line[2]
            weight = int(line[3].split(':')[1]) 
            graph.add_edge(start_vertex, end_vertex, weight=weight)

    return graph



import random
from Bio import SeqIO

# 读取FASTA文件并获取基因组序列
def read_genome(fasta_file):
    genome = SeqIO.read(fasta_file, "fasta")
    return genome.seq

# 随机改变指定比例的碱基
def random_change(genome, change_ratio):
    genome_list = list(genome)  # 转换为列表，便于修改
    total_bases = len(genome_list)
    num_changes = int(total_bases * change_ratio)  # 计算需要改变的碱基数量

    for _ in range(num_changes):
        position = random.randint(0, total_bases - 1)  # 随机选择一个位置
        original_base = genome_list[position]
        bases = ['A', 'T', 'C', 'G']
        bases.remove(original_base)  # 确保选择不同的碱基
        new_base = random.choice(bases)  # 随机选择新的碱基
        genome_list[position] = new_base  # 替换为新的碱基

    return "".join(genome_list)

# 模拟基因组的随机改变
def simulate_random_changes(input_fasta, output_fasta, change_ratio):
    # 读取基因组
    genome = read_genome(input_fasta)

    # 执行随机改变
    mutated_genome = random_change(genome, change_ratio)

    # 保存变异后的基因组
    with open(output_fasta, "w") as out_f:
        out_f.write(f">mutated_genome\n{mutated_genome}")

    print(f"Original Genome Length: {len(genome)}")
    print(f"Mutated Genome Length: {len(mutated_genome)}")
    print(f"Change Ratio: {change_ratio * 100}%")
    print(f"Number of changes: {int(len(genome) * change_ratio)}")

# 主程序，指定输入和输出路径
input_fasta = "JRCSF.fa"  # 输入的基因组FASTA文件路径
output_fasta = "30_jrcsf.fasta"  # 输出的突变基因组FASTA文件路径
change_ratio = 0.30  # 改变的比例，例如15%

simulate_random_changes(input_fasta, output_fasta, change_ratio)


import networkx as nx
from . import readgraph as rg
from . import choosepaths as cp
from . import writesequence as ws
import os
import argparse
import utils
from . import polisher as pl

import edit_distance as ed
from networkx.readwrite.graph6 import write_graph6
import shutil
import os
import time
import datetime

def main():
	parser = argparse.ArgumentParser(description="Perform graph-based sequence assembly.")
	parser.add_argument('-r','--reads',type=str, required=True, help="Reads file for graph construction (in fasta format).")
	parser.add_argument('-b', '--backbone',type=str, required=True, help="Backbone sequence file for graph construction (in fasta format).")
	parser.add_argument('-n', '--iterations', type=int, required=False, default=15 , help="Maximum Number of Iterations for graph construction.")
	parser.add_argument('-od','--output_dir', type=str, required=True, help="Outdir")

	args = parser.parse_args()
	ec_reads = args.reads
	backbone = args.backbone
	# groundtruth = args.gt
	num_iterations = args.iterations
	outdir = args.output_dir

	graph_out_pref = os.path.join(ec_reads + "_ON_" + os.path.splitext(os.path.basename(backbone))[0])
	graph_out_pref = os.path.basename(graph_out_pref)


	for i in range(num_iterations):

		graph_filename = f"iter{i+1}.graph"
		graph_folder = os.path.join(graph_out_pref, f"iter{i+1}")
		graph_location = os.path.join(graph_folder, graph_filename)

		os.makedirs(graph_folder, exist_ok=True)
	


		start_time = time.perf_counter()
		oldbackbone = backbone

		# print(f"Function execution time: {end_time - start_time:.6f} seconds")
		if i == 0 : 
			aln_graph = utils.construct_aln_graph_from_fasta(ec_reads, backbone)
			aln_graph.merge_nodes()
		else :
			aln_graph = utils.construct_aln_graph_from_fasta(ec_reads, pb)
			aln_graph.merge_nodes()
		# aln_graph = utils.construct_aln_graph_from_fasta(ec_reads, consensus_location)

		# aln_graph.merge_nodes()
		# pbdconsensus = aln_graph.generate_consensus()[0]
	
		# pbdconsensus_location = os.path.join( os.path.join(graph_folder , f"pbdconsensus"), 'output_{}_{}.fa'.format(f"iter{i+1}", 0))
	
		nx_pack = utils.aln2nx(aln_graph)

		utils.nx2gfa(nx_pack[0], graph_filename, graph_folder)

		# 读取图文件
		graph = rg.read_graph_from_file(graph_location)
		print(graph)

		topological_order = list(nx.topological_sort(graph))
		start_node = topological_order[0]
		start_nodes = [node for node in graph.nodes() if graph.in_degree(node) == 0]
		end_nodes = [node for node in graph.nodes() if graph.out_degree(node) == 0]
		end_node = topological_order[-1]
	
		# 计算最大权重路径
		# max_weight, max_weight_path, max_weight_sequences, node_graph = cp.max_weight_path_cuda(graph)
		max_weight, max_weight_path, max_weight_sequences, node_graph = cp.max_weight_path_old(graph, start_node, end_node)
		# max_weight, max_weight_path, max_weight_sequences, node_graph = cp.max_weight_path_gpu(graph, start_node, end_node)

		end_time = time.perf_counter()

		polished_graph, polished_sequences = pl.polish(node_graph)
		pb=ws.store_labels_as_fa_hanshuming(polished_sequences, graph_folder,'dp_polish')
		
		backbone = graph_location = os.path.join(graph_folder, 'output_{}_{}.fa'.format(graph_out_pref, 0))
		print("graph_folder", graph_folder)
		sequence_location = ws.store_labels_as_fa(max_weight_sequences, graph_folder)
		# pbdconsensus =  ws.store_labels_as_fa([pbdconsensus[::-1]], os.path.join(graph_folder , f"pbdconsensus"))
		print(f"Iteration {i+1} completed.")
		print("-----------------------------------------------------------------------------------------------")
		if utils.are_sequences_identical(oldbackbone, sequence_location):
			break
		
		sequence_location = os.path.join(graph_folder, 'output_{}_{}.fa'.format(graph_out_pref, 0))
		if i == 1:
			 dp_noiter_result = pb


		
	
	# dp_time = end_time - start_time
	#os.system(f"quast.py {backbone} -r {groundtruth}")
	polish_folder = os.path.join(graph_out_pref, f"polish")
	os.makedirs(polish_folder, exist_ok=True)
	polished_location = os.path.join(polish_folder, 'output_{}_{}.fa'.format(graph_out_pref, 0))



	polished_graph, polished_sequences = pl.polish(node_graph)
	# nx.write_gml(node_graph, os.path.join(polish_folder, 'graph.gml'))
	# nx.write_gml(polished_graph, os.path.join(polish_folder, 'polished_graph.gml'))

	ws.store_labels_as_fa(polished_sequences, polish_folder)

	if not os.path.exists(outdir):
		os.makedirs(outdir)
	# move together
	dst_file=os.path.join(outdir, 'pvga_{}.fa'.format(graph_out_pref))
	shutil.copy(polished_location, dst_file)

	# dst_file=os.path.join(outdir, "dp_result.fa")
	# shutil.copy(sequence_location, dst_file)

	dst_file=os.path.join(outdir, 'pvga_noiter_{}.fa'.format(graph_out_pref))
	shutil.copy(dp_noiter_result, dst_file)

	# dst_file=os.path.join(outdir, "vote_consensus.fa")
	# shutil.copy(consensus_location, dst_file)

	# import writegraphconsensus as wss
	# wss.store_labels_as_fa([graph_consensus], polish_folder)

# 构图
if __name__ == "__main__":
	main()


import sys
import pysam
import numpy as np
import os
from Bio import SeqIO

truth = sys.argv[1]
query = sys.argv[2]


os.system("rm -rf alignment_file")
os.system("mkdir alignment_file")
truth_name = os.path.basename(os.path.dirname(truth))
query_name = os.path.basename(os.path.dirname(query))

samfile = f"alignment_file/{query_name}_on_{truth_name}.sam"
unique_samfile = f"alignment_file/{query_name}_on_{truth_name}_unique.sam"
bamfile = f"alignment_file/{query_name}_on_{truth_name}_unique.bam"
sorted_bamfile = f"alignment_file/{query_name}_on_{truth_name}_unique_sorted.bam"


os.system(f'minimap2 -a {truth} {query} > {samfile}')
os.system(f"samtools view -h -F 0x900 {samfile} > {unique_samfile}")
os.system(f'samtools view -b {unique_samfile} > {bamfile}')
os.system(f'samtools sort {bamfile} -o {sorted_bamfile}')
os.system(f'samtools index {sorted_bamfile}')


file_bam = sorted_bamfile
bamfile = pysam.AlignmentFile(file_bam)

for s in SeqIO.parse(query,'fasta'):
    query_length = len(s.seq)
    print(f"query_length:{query_length}")
    
for s in SeqIO.parse(truth,'fasta'):
    truth_length = len(s.seq)
    print(f"truth_length:{truth_length}")



contig_length = []
ref_length = []
cigar_length = []
mismatch = []
indel = []
edit_distance = []
contig_name = []
contig_ref = []
for contig in bamfile.fetch():
    align_length = contig.reference_length
    cigar = contig.get_cigar_stats()[0]
    indel = cigar[1]+cigar[2]
    unaligned = cigar[4]
    diff = cigar[-1]
    mismatch = diff-indel
    edit_dist = unaligned + diff + truth_length - align_length
    print(f"unaligned_length:{unaligned}")
    print(f"edit_distance:{edit_dist}")
    print(f"total_mismatch:{mismatch}")
    print(f"total_indel:{indel}")


import sys
import pysam
import numpy as np
import os
from Bio import SeqIO

def calculate_alignment_stats(query, truth):

    os.system("rm -rf alignment_file")
    os.system("mkdir alignment_file")
    truth_name = os.path.basename(os.path.dirname(truth))
    query_name = os.path.basename(os.path.dirname(query))

    samfile = f"alignment_file/{query_name}_on_{truth_name}.sam"
    unique_samfile = f"alignment_file/{query_name}_on_{truth_name}_unique.sam"
    bamfile = f"alignment_file/{query_name}_on_{truth_name}_unique.bam"
    sorted_bamfile = f"alignment_file/{query_name}_on_{truth_name}_unique_sorted.bam"

    os.system(f'minimap2 -a {truth} {query} > {samfile}')
    os.system(f"samtools view -h -F 0x900 {samfile} > {unique_samfile}")
    os.system(f'samtools view -b {unique_samfile} > {bamfile}')
    os.system(f'samtools sort {bamfile} -o {sorted_bamfile}')
    os.system(f'samtools index {sorted_bamfile}')

    file_bam = sorted_bamfile
    bamfile = pysam.AlignmentFile(file_bam)
    
    for s in SeqIO.parse(query, 'fasta'):
        query_length = len(s.seq)
        print({query_name} ,":", {query_length})

    for s in SeqIO.parse(truth, 'fasta'):
        truth_length = len(s.seq)
        print({truth_name} ,":", {truth_length})

    contig_length = []
    ref_length = []
    cigar_length = []
    mismatch = []
    indel = []
    edit_distance = []
    contig_name = []
    contig_ref = []

    for contig in bamfile.fetch():
        align_length = contig.reference_length
        cigar = contig.get_cigar_stats()[0]
        indel = cigar[1] + cigar[2]
        unaligned = cigar[4]
        diff = cigar[-1]
        mismatch = diff - indel
        edit_dist = unaligned + diff + truth_length - align_length
        print(f"unaligned_length: {unaligned}")
        print(f"edit_distance: {edit_dist}")
        print(f"total_mismatch: {mismatch}")
        print(f"total_indel: {indel}")



import sys
import numpy as np
from Bio import SeqIO
file_in = sys.argv[1]
f = open(file_in,'r')
seq = []
name = []

for s in SeqIO.parse(file_in,'fasta'):
    name.append(s.description)
    seq.append([c for c in s.seq])

# while True:
#     line = f.readline().strip()
#     if not line:
#         break
#     seq.append([c for c in f.readline().strip()])
#     name.append(line)

print(name)

l = len(name)
seq = np.array(seq)
flag = []
for s in seq:
	flag.append(s!='-')

# seq_type = {'896':0,'HXB2':1,'JRCSF':2,'NL43':3,'YU2':4}


length_all = np.zeros([l,l])
length2 = np.zeros([l,l])
div_mutation = np.zeros([l,l])
div_all = np.zeros([l,l])
for i in range(l-1):
	for j in range(i+1,l):
		diff = seq[i]!=seq[j]
		flag_ij = flag[i]*flag[j]
		all = np.count_nonzero(diff)
		mut = np.count_nonzero(diff*flag_ij)
		div_mutation[i,j] = mut
		div_all[i,j] = all
		length_all[i,j] = np.count_nonzero(flag[i]+flag[j])
		length2[i,j] = np.count_nonzero(flag[i]*flag[j])

div_mutation = div_mutation + div_mutation.T
div_all = div_all + div_all.T
length_all = length_all + length_all.T
length2 = length2 + length2.T
# print(div_mutation)
# print(div_all)
# print(len(seq[0]))
# print(length_all)
# print(np.round(div_all/(length_all+1e-16),4))
print((1-np.round(div_all/(length_all+1e-16),4)))



diversity_all = 100-np.round(div_all/(length_all+1e-16),4)*100


from matplotlib import pyplot as plt
# alpha = [i for i in seq_type.keys()]
alpha = [i for i in name]

fig = plt.figure()
ax = fig.add_subplot(111)
cax = ax.matshow(diversity_all, interpolation='nearest')
for (x, y), value in np.ndenumerate(diversity_all):
    plt.text(x, y, f"{value:.2f}", va="center", ha="center")

fig.colorbar(cax)

xaxis = np.arange(len(alpha))
ax.set_xticks(xaxis)
ax.set_yticks(xaxis)
ax.set_xticklabels(alpha)
ax.set_yticklabels(alpha)
plt.show()



from setuptools import setup, find_packages

setup(
    name='pvga',  # 确保包名有效
    version='0.0.3',
    description='PVGA is a powerful virus-focused assembler that does both assembly and polishing.',
    author='zhisong',
    author_email='songzhics@gmail.com',
    url='https://github.com/SoSongzhi/PVGA',
    
    # 使用 find_packages 自动发现包
    package_dir={"": "src"},  # 指定根目录为 src
    packages=find_packages(where="src"),  # 自动发现 src 下的包
    
    # 使用 entry_points 替代 scripts
    entry_points={
        "console_scripts": [
            "pvga=pvga:main"  # 定义命令行工具
        ]
    },
    
    # 其他可选配置
    python_requires='>=3.7',  # 指定 Python 版本要求
    install_requires=[  # 添加依赖项
        # 'numpy>=1.20.0',
        # 'pandas>=1.3.0',
    ],
)


import re
import textwrap
import collections
import time
import operator
from random import shuffle

rev_table = {"A": "T", "T": "A", "C": "G", "G": "C", "-": "-"}


def reverse_graph(aln_graph):
    """
    reverse the graph
    :param aln_graph: graph we reverse
    :return:
    """
    graph_list = aln_graph.get_sorted_nodes()

    origin_nodes_to_edge = aln_graph.nodes_to_edge
    in_node_list = {}
    for node in graph_list:

        if node.base == "B":
            node.base = "E"

        elif node.base == "E":
            node.base = "B"

        else:
            node.base = rev_table[node.base]

        in_node_list[node] = []
        for in_edge in node._in_edges:
            in_node = in_edge.in_node
            in_node_list[node].append(in_node)

    del_list = []
    for edge in aln_graph.edges:
        del_list.append(aln_graph.edges[edge])

    for edge in del_list:
        aln_graph.delete_edge(edge)

    for node in graph_list:

        for in_node in in_node_list[node]:
            edge_add = AlnEdge(node, in_node)
            edge_add.count = origin_nodes_to_edge[(in_node.ID, node.ID)].count
            edge_add.score = origin_nodes_to_edge[(in_node.ID, node.ID)].score
            aln_graph.add_edge(edge_add)
    
    node_temp = aln_graph.begin_node
    aln_graph.begin_node = aln_graph.end_node
    aln_graph.end_node = node_temp
    
        
    rev_back_bone_nodes = {}
    forward_keys = list(aln_graph.backbone_nodes.keys())        
    for node_index, index in enumerate(reversed(forward_keys)):
        new_key = forward_keys[index]
        rev_back_bone_nodes[new_key] = aln_graph.backbone_nodes[node_index]
            
    aln_graph.backbone_nodes = rev_back_bone_nodes
    
    old_start = aln_graph.consensus_start
    old_end = aln_graph.consensus_end
    if aln_graph.consensus_start != None:
        aln_graph.consensus_start = forward_keys[-1] - old_end
    if aln_graph.consensus_end != None:
        aln_graph.consensus_end = forward_keys[-1] - old_start
    # print aln_graph.consensus_start, aln_graph.consensus_end, forward_keys[-1]

def convert_mismatches(alnQ, alnT):
    qs = []
    ts = []
    assert len(alnQ) == len(alnT)
    alnQ = list(alnQ)
    alnT = list(alnT)
    for i in range(len(alnQ)):
        qb = alnQ[i]
        tb = alnT[i]
        if qb != tb and qb != "-" and tb != "-":
            qs.append("-")
            ts.append(tb)
            qs.append(qb)
            ts.append("-")
        else:
            qs.append(qb)
            ts.append(tb)
    alnQ = qs
    alnT = ts

    # push gaps to the right
    for i in range(len(alnQ) - 1):
        if alnT[i] == "-":
            j = i
            while 1:
                j += 1
                c = alnT[j]
                if c != "-" or j >= len(alnQ) - 1:
                    break
            if c == alnQ[i]:
                alnT[i] = c
                alnT[j] = "-"

        if alnQ[i] == "-":
            j = i
            while 1:
                j += 1
                c = alnQ[j]
                if c != "-" or j >= len(alnT) - 1:
                    break
            if c == alnT[i]:
                alnQ[i] = c
                alnQ[j] = "-"
    qs2 = []
    ts2 = []
    for i in range(len(alnQ)):
        if alnQ[i] != '-' or alnT[i] != '-':
            qs2.append(alnQ[i])
            ts2.append(alnT[i])

    return "".join(qs2), "".join(ts2)


class AlnEdge(object):
    def __init__(self, in_node, out_node):
        self.ID = id(self)
        self.in_node = in_node
        self.out_node = out_node
        in_node.addout_edge(self)
        out_node.add_in_edge(self)
        self.visited = False
        self.count = 0
        self.score = 0

    def increase_count(self):
        self.count += 1

    def set_score(self, s):
        self.score = s

    def add_to_score(self, s):
        self.score += s

    def __repr__(self):
        return "(edge_ID:%d, in_node:%s, out_node:%s)" % (self.ID, self.in_node.__repr__(), self.out_node.__repr__())


class AlnNode(object):
    def __init__(self, base):
        self.ID = id(self) # inbuilt id() function; 
        self.base = base
        self._in_edges = []
        self._out_edges = []
        self.is_backbone = False
        self.coverage = 0
        self.weight = 0
        self.backbone_node = None
        self.best_in_edge = None
        self.best_out_edge = None
        self.info = []

    def add_in_edge(self, in_edge):
        self._in_edges.append(in_edge)

    def addout_edge(self, out_edge):
        self._out_edges.append(out_edge)

    def increase_weight(self, w=1):
        self.weight += w

    def in_edges(self):
        return self._in_edges

    def __repr__(self):
        return "(node_id:%d, base:%s, b:%s, w:%d, c:%d)" % (
            self.ID, self.base, self.is_backbone, self.weight, self.coverage)


class AlnGraph(object):
    def __init__(self, backbone_seq):
        """
        AlnGraph is instantiated by giving a "backbone" sequence
        >>> g = AlnGraph("ACGTCTAT")
        """

        self.nodes = {}
        self.edges = {}
        self.backbone_nodes = {}
        self.backbone_to_reads = {}
        self.nodes_to_edge = {}
        self._max_node_id = 0
        self._max_edge_id = 0
        self.consensus_path = None
        self.consensus_str = None
        self.consensus_start = None
        self.consensus_end = None
        self.path_count = 0

        self.begin_node = AlnNode("B")
        self.begin_node.backbone_node = self.begin_node
        self.add_node(self.begin_node)
        self.read_range = {}
        last_node = self.begin_node

        # set up a barebone linear graph for the seed sequence
        for pos in range(len(backbone_seq)):
            node = AlnNode(backbone_seq[pos])
            self.backbone_nodes[pos] = node
            node.is_backbone = True
            node.backbone_node = node
            node.increase_weight()
            self.add_node(node)

            edge = AlnEdge(last_node, node)
            edge.set_score(0)
            edge.increase_count()
            self.add_edge(edge)
            last_node = node

        self.end_node = AlnNode("E")
        self.end_node.backbone_node = self.end_node
        self.add_node(self.end_node)
        edge = AlnEdge(last_node, self.end_node)
        edge.set_score(0)
        edge.increase_count()
        self.add_edge(edge)

        self.backbone_node_to_pos = dict(list(zip(list(self.backbone_nodes.values()),
                                             list(self.backbone_nodes.keys()))))

    def add_alignment(self, aln, rId=None):

        read_pos, read_end_pos, read_aln_seq = aln[0]
        backbone_pos, backbone_end_pos, backbone_aln_seq = aln[1]

        if rId != None:
            self.read_range[rId] = (backbone_pos, backbone_end_pos)

        last_node = self.begin_node

        for aln_pos in range(len(read_aln_seq)):
            read_base = read_aln_seq[aln_pos]
            backbone_base = backbone_aln_seq[aln_pos]

            if read_base == backbone_base:
                node = self.backbone_nodes[backbone_pos]
                node.increase_weight()
                node.backbone_node.coverage += 1

                if (last_node.ID, node.ID) not in self.nodes_to_edge:
                    edge = AlnEdge(last_node, node)
                    self.add_edge(edge)
                    edge.increase_count()
                else:
                    self.nodes_to_edge[(last_node.ID, node.ID)].increase_count()

                read_pos += 1
                backbone_pos += 1
                last_node = node
                if rId != None:
                    node.info.append(rId)

            elif read_base == "-" and backbone_base != "-":
                node = self.backbone_nodes[backbone_pos]
                node.backbone_node.coverage += 1
                backbone_pos += 1

            elif read_base != "-" and backbone_base == "-":
                node = AlnNode(read_base)
                node.increase_weight()
                node.backbone_node = self.backbone_nodes[backbone_pos]

                self.add_node(node)

                edge = AlnEdge(last_node, node)
                self.add_edge(edge)
                edge.increase_count()

                read_pos += 1
                last_node = node
                if rId != None:
                    node.info.append(rId)

            self.backbone_to_reads.setdefault(backbone_pos, set())

            if rId != None:
                self.backbone_to_reads[backbone_pos].add(rId)

        if (last_node.ID, self.end_node.ID) not in self.nodes_to_edge:
            edge = AlnEdge(last_node, self.end_node)
            edge.set_score(0)
            edge.increase_count()
            self.add_edge(edge)

    def add_edge(self, edge):
        edge.ID = id(edge)
        self.edges[edge.ID] = edge
        self.nodes_to_edge[(edge.in_node.ID, edge.out_node.ID)] = edge

    def add_node(self, node):
        self.nodes[node.ID] = node

    def delete_edge(self, edge):
        n1 = edge.in_node
        n2 = edge.out_node
        n1._out_edges.remove(edge)
        n2._in_edges.remove(edge)
        del self.edges[edge.ID]
        del edge

    def delete_node(self, node):
        for in_edge in node._in_edges:
            self.delete_edge(in_edge)
        for out_edge in node._out_edges:
            self.delete_edge(out_edge)
        del self.nodes[node.ID]
        del node

    def merge_in_nodes(self, node):

        node_groups = {}
        for in_edge in node._in_edges:
            in_node = in_edge.in_node
            if len(in_node._out_edges) == 1:
                node_groups.setdefault(in_node.base, [])
                node_groups[in_node.base].append(in_node)

        for b in node_groups:
            if len(node_groups[b]) <= 1:
                continue

            nodes = node_groups[b]

            for n in nodes[1:]:
                nodes[0]._out_edges[0].count += n._out_edges[0].count
                nodes[0].increase_weight(n.weight)
                nodes[0].info.extend(n.info)

            for n in nodes[1:]:
                for in_edge in n._in_edges:
                    n1 = in_edge.in_node
                    if (n1.ID, nodes[0].ID) not in self.nodes_to_edge:
                        e = AlnEdge(n1, nodes[0])
                        self.add_edge(e)
                        e.count = self.nodes_to_edge[(n1.ID, n.ID)].count
                        e.visited = self.nodes_to_edge[(n1.ID, n.ID)].visited
                    else:
                        e = self.nodes_to_edge[(n1.ID, nodes[0].ID)]
                        e.count += self.nodes_to_edge[(n1.ID, n.ID)].count
                        e.add_to_score(self.nodes_to_edge[(n1.ID, n.ID)].score)
                self.delete_node(n)

            self.merge_in_nodes(nodes[0])

    def merge_out_nodes(self, node):

        node_groups = {}

        for out_edge in node._out_edges:
            out_node = out_edge.out_node
            if len(out_node._in_edges) == 1:
                node_groups.setdefault(out_node.base, [])
                node_groups[out_node.base].append(out_node)

        for b in node_groups:
            if len(node_groups[b]) <= 1:
                continue

            nodes = node_groups[b]

            for n in nodes[1:]:
                nodes[0]._in_edges[0].count += n._in_edges[0].count
                nodes[0].increase_weight(n.weight)
                nodes[0].info.extend(n.info)

            for n in nodes[1:]:
                for out_edge in n._out_edges:
                    n2 = out_edge.out_node
                    if (nodes[0].ID, n2.ID) not in self.nodes_to_edge:
                        e = AlnEdge(nodes[0], n2)
                        self.add_edge(e)
                        e.count = self.nodes_to_edge[(n.ID, n2.ID)].count
                        e.visited = self.nodes_to_edge[(n.ID, n2.ID)].visited
                    else:
                        e = self.nodes_to_edge[(nodes[0].ID, n2.ID)]
                        e.count += self.nodes_to_edge[(n.ID, n2.ID)].count
                        e.add_to_score(self.nodes_to_edge[(n.ID, n2.ID)].score)

                self.delete_node(n)

    def merge_nodes(self):
        for e in list(self.edges.values()):
            e.visited = False

        seed_nodes = []

        for node_id, node in list(self.nodes.items()):
            if len(node._in_edges) == 0:
                seed_nodes.append(node)

        while 1:
            if len(seed_nodes) == 0:
                break
            node = seed_nodes.pop(0)
            self.merge_in_nodes(node)
            self.merge_out_nodes(node)

            for out_edge in node._out_edges:
                out_node = out_edge.out_node
                out_edge.visited = True
                in_edges = [e for e in out_node._in_edges if e.visited == False]
                if len(in_edges) == 0:
                    seed_nodes.append(out_node)

    def find_best_path(self):
        seed_nodes = []
        node_score = {}
        best_node_score_edge = {}

        for node_id, node in list(self.nodes.items()):
            if len(node._out_edges) == 0:
                seed_nodes.append(node)
                node_score[node.ID] = 0
                node.best_out_edge = None
                node.best_in_edge = None

        for edge in list(self.edges.values()):
            edge.visited = False

        while 1:

            if len(seed_nodes) == 0:
                break

            node = seed_nodes.pop(0)
            best_edge = None
            best_score = None
            for out_edge in node._out_edges:
                out_node = out_edge.out_node

                score = node_score[out_node.ID]

                backbone_node_coverage = out_node.backbone_node.coverage
        
                if out_node.is_backbone == True and out_node.weight == 1:
                    new_score = score - 10
                else:
                    new_score = out_edge.count - backbone_node_coverage * 0.5 + score
                if best_score == None or new_score > best_score:
                    best_edge = out_edge
                    best_score = new_score

            if best_edge != None:
                best_node_score_edge[node.ID] = best_edge
                node_score[node.ID] = best_score

            for in_edge in node._in_edges:
                in_node = in_edge.in_node
                in_edge.visited = True
                out_edges = [e for e in in_node._out_edges if e.visited == False]
                if len(out_edges) == 0:
                    seed_nodes.append(in_node)

        node = self.begin_node
        consensus_path = []
        
        con_start = len(self.backbone_nodes)
        con_end = 0

        while 1:
            consensus_path.append(node)
            if node.ID not in best_node_score_edge:
                break
            else:
                best_out_edge = best_node_score_edge[node.ID]
                node.best_out_edge = best_out_edge
                best_node_score_edge[node.ID].out_node.best_in_edge = best_node_score_edge[node.ID]
                node = best_node_score_edge[node.ID].out_node
                node.best_in_edge = best_out_edge

        # print consensus_path
        con_start = len(self.backbone_nodes)
        con_end = 0
        
        for node in self.backbone_nodes:
            # print self.backbone_nodes[node]
            if self.backbone_nodes[node] in consensus_path:
                # print "hhaa"
                if int(node) < con_start:
                    con_start = node
                if int(node) > con_end: 
                    con_end = node

        self.consensus_start = con_start
        self.consensus_end = con_end
        # print self.consensus_start
        # print self.consensus_end
        print("path score of consensus_path:", self.score(consensus_path))
        return consensus_path

    def generate_consensus(self, min_cov=0):

        self.merge_nodes()

        if self.consensus_path == None:
            self.consensus_path = self.find_best_path()

        s = []
        c = []
        cov = []    # a list for the coverage of all node in the path
        cov_good = []
        for n in self.consensus_path:
            if n not in [self.begin_node, self.end_node]:
                s.append(n.base)
                cov.append(n.backbone_node.coverage)
                if len(n.info) >= min_cov:
                    # print len(n.info), n.info
                    cov_good.append("1")
                else:
                    cov_good.append("0")

                rn = n.weight
                if n.is_backbone == True:
                    rn -= 1
                if n.best_out_edge != None:
                    en = n.best_out_edge.count
                else:
                    en = 0
                if n.best_in_edge != None:
                    en2 = n.best_in_edge.count
                else:
                    en2 = 0
                c.append((rn, en2, en))

        s = "".join(s)
        cov_good = "".join(cov_good)

        p = re.compile("1+")
        best_range = (0, 0)
        best_l = 0
        for m in p.finditer(cov_good):
            b, e = m.start(), m.end()
            if e - b > best_l:
                best_range = (b, e)
                best_l = e - b
        b, e = best_range
        # print "best range", b, e
        if e - b != 0:
            self.consensus_str = s[b:e]
            c = c[b:e]
            cov = cov[b:e]
        else:
            self.consensus_str = ""
            c = []

        return self.consensus_str, c, cov, self.consensus_start, self.consensus_end

    def get_high_entropy_nodes(self,
                               ignore_backbone=False,
                               coverage_th=20,
                               overlap_th=10,
                               entropy_th=0.5):
        from math import log

        if self.consensus_path == None:
            self.generate_consensus()

        backbone_node_to_pos = self.backbone_node_to_pos

        node_entropy = []

        for node_id in self.nodes:

            node = self.nodes[node_id]

            if node.backbone_node.coverage < coverage_th: continue
            if ignore_backbone and node.is_backbone: continue

            p = 1.0 * (len(node.info) + 1) / (node.backbone_node.coverage + 1)
            if abs(p - 1) < 1e-5 or p > 1:
                ent = 0
            else:
                ent = - p * log(p) - (1 - p) * log(1 - p)

            node_entropy.append([node_id, node, ent])

        node_entropy.sort(key=lambda x: -x[2])

        high_entropy_nodes = [n for n in node_entropy if n[2] > entropy_th]

        return node_entropy, high_entropy_nodes

    def get_sorted_nodes(self):
        seed_nodes = []
        # first loop should only add the "Start" node into seed_nodes
        for node_id, node in list(self.nodes.items()):
            if len(node._in_edges) == 0:
                seed_nodes.append(node)

        # all flagged as unvisited
        for edge in list(self.edges.values()):
            edge.visited = False

        sorted_nodes = []

        while 1:
            if len(seed_nodes) == 0:
                break

            # pop(index) removes the item at the given index from the list and returns the removed item.
            node = seed_nodes.pop(0)

            for in_edge in node._in_edges:
                in_node = in_edge.in_node
            # why this step? in_node not used
            sorted_nodes.append(node)

            for out_edge in node._out_edges:
                out_node = out_edge.out_node
                out_edge.visited = True
                in_edges = [e for e in out_node._in_edges if e.visited == False]
                if len(in_edges) == 0:
                    seed_nodes.append(out_node)

        return sorted_nodes

    def get_edge_count(self, src, des):
        find = False
        for edge in src._out_edges:
            if edge.out_node == des:
                flag = True
                return edge.count
        if not find:
            return 0
    
    def score(self, res):
        '''
        sum_score = 0
        for i in range(len(res)-2):
            sum_score += self.get_edge_count(res[i], res[i+1])
        return sum_score
        '''
        return sum([self.get_edge_count(res[i],res[i+1]) for i in range(len(res)-2)])

    def all_topo_sort_util(self, res, nodes_visited, src_node, des_node, out):
        '''
        node_dic = self.nodes
        for node_id, node in node_dic.items():
        '''
        # mark current src node as visited and store in path
        nodes_visited[src_node.ID] == True
        res.append(src_node)
        if src_node == des_node:
            if self.path_count >= 1000:
                return
            else:
                #print(">score_", self.score(res))
                base_path = [node.base for node in res]
                self.path_count += 1
                print(''.join(base_path[1:-1]), self.score(res), file=out)
            
        else:
            out_degree = sum([out_e.count for out_e in list(src_node._out_edges)])
            out_nodes_l = [out_e.out_node for out_e in list(src_node._out_edges)] 
            selected_nodes_l =[]           
            
            # only traverse to path that has weight higher than 1/(number of out edges):
            '''for out_e in src_node._out_edges:
                if float(out_e.count/out_degree) >= 1/len(out_nodes_l):
                    selected_nodes_l.append(out_e.out_node)
            #for next_node in out_nodes_l:
            for next_node in selected_nodes_l:
                if nodes_visited[next_node.ID] == False:
                    self.all_topo_sort_util(res, nodes_visited, next_node, des_node, out)    
            '''            
            # !! only traverse to path that has weight higher than 1/2 will fail to when node is diverged

            # sort the out_edge first and traverse from the most possible node
            out_edges_sorted = sorted(src_node._out_edges, key=operator.attrgetter('count'))
            # selected_nodes_l.append(out_edges_sorted[0].out_node)
            for out_e in out_edges_sorted:
                next_node = out_e.out_node
                if nodes_visited[next_node.ID] == False:
                    self.all_topo_sort_util(res, nodes_visited, next_node, des_node, out)
            
        res.pop()
        nodes_visited[src_node.ID] == False


    # call the all_topo_sort_util() recursively
    def all_topo_sort(self,out):
        '''
        edges_cp = self.edges.values()
        for edge in list(edges_cp):
            edge.visited = False
        '''
        s_t = time.time()
        nodes_visited = {node_id: False for node_id in list(self.nodes.keys())}
        res = []
        with open(out+"_path.fa", 'w+') as path_out:
            self.all_topo_sort_util(res, nodes_visited, self.begin_node, self.end_node, path_out)
        print("there are %d path found in the graph"%(self.path_count))
        print("Taken time: %s seconds"%(time.time() - s_t))

    def random_walk_util(self, res, nodes_visited, src_node, des_node, out):
        '''
        node_dic = self.nodes
        for node_id, node in node_dic.items():
        '''
        # mark current src node as visited and store in path
        nodes_visited[src_node.ID] == True
        res.append(src_node)
        if self.path_count >= 2000:
            return
        if src_node == des_node:
            if self.path_count >= 2000:
                return
            else:
                #print(">score_", self.score(res))
                base_path = [node.base for node in res]
                self.path_count += 1
                print(''.join(base_path[1:-1]), self.score(res), file=out)
            
        else:
            out_nodes_l = [out_e.out_node for out_e in list(src_node._out_edges)] 
            shuffle(out_nodes_l)       
            for next_node in out_nodes_l:
                if nodes_visited[next_node.ID] == False:
                    self.all_topo_sort_util(res, nodes_visited, next_node, des_node, out)            
        res.pop()
        nodes_visited[src_node.ID] == False

    def random_walk_k(self):
        k = 1000
        with open("random_walk_k.fa", 'w') as path_out:
            for i in range(k):
                rd_path = []
                rd_path.append(self.begin_node)
                while(rd_path[-1].base != 'E'):
                    out_nodes_l = [out_e.out_node for out_e in list(rd_path[-1]._out_edges)] 
                    shuffle(out_nodes_l) 
                    rd_path.append(out_nodes_l[0])
                base_path = [node.base for node in rd_path]
                print(''.join(base_path[1:-1]), self.score(rd_path), file=path_out)
            print("finish %d random walking"%i)
        return 0

    # call the all_topo_sort_util() recursively
    def random_walk(self,out):
        '''
        edges_cp = self.edges.values()
        for edge in list(edges_cp):
            edge.visited = False
        '''
        out_N = 1000
        s_t = time.time()
        nodes_visited = {node_id: False for node_id in list(self.nodes.keys())}
        res = []
        with open(out+"_random_path.fa", 'w+') as path_out:
            self.random_walk_util(res, nodes_visited, self.begin_node, self.end_node, path_out)
        print("there are %d path found in the graph"%(self.path_count))
        print("Taken time: %s seconds"%(time.time() - s_t))

    def get_read_node_vector(self, entropy_th=0.5):

        ne, hne = self.get_high_entropy_nodes(coverage_th=0, entropy_th=entropy_th)
        node_to_entropy = dict([(v[1], v[2]) for v in ne])
        read_ids = set()

        for n in list(self.nodes.values()):
            for r in n.info:
                read_ids.add(r)

        backbone_node_to_pos = self.backbone_node_to_pos

        sn = self.get_sorted_nodes()
        high_entropy_nodes = [(n,
                               backbone_node_to_pos[n.backbone_node],
                               node_to_entropy[n]) for n in sn if node_to_entropy[n] > entropy_th]

        read_to_nodes = {}
        for r_id in read_ids:
            read_to_nodes[r_id] = ["-"] * len(high_entropy_nodes)
            for i, n in enumerate(high_entropy_nodes):
                if r_id in n[0].info:
                    read_to_nodes[r_id][i] = n[0].base

        return read_to_nodes, high_entropy_nodes

    def output_consensus_fasta(self, f_name, rID):
        f = open(f_name, "a")
        print(">%s" % rID, file=f)
        for i in range(0, len(self.consensus_str), 60):
            print(self.consensus_str[i:i + 60], file=f)
        f.close()

    def filter_graph(self, threshold = 8):
        edge_to_remove = []
        for edge in self.edges:
            in_node = self.edges[edge].in_node
            out_node = self.edges[edge].out_node

            if in_node.backbone_node.coverage < out_node.backbone_node.coverage:
                coverage = in_node.backbone_node.coverage
            else:
                coverage = out_node.backbone_node.coverage
    
            if self.edges[edge].count <= 1 and coverage > threshold:
                edge_to_remove.append(self.edges[edge])

        for edge in edge_to_remove:
            self.delete_edge(edge)
        
        node_to_remove = []
        #  and (self.nodes[node].is_backbone == False)
        for node in self.nodes:
            if (not(self.nodes[node]._in_edges) and not(self.nodes[node]._out_edges)):
                node_to_remove.append(self.nodes[node])
            elif not(self.nodes[node]._in_edges) and (self.nodes[node] != self.begin_node):
                edge_add = AlnEdge(self.begin_node, self.nodes[node])
                edge_add.count = 1
                edge_add.score = 0
                self.add_edge(edge_add)
            elif not(self.nodes[node]._out_edges) and (self.nodes[node] != self.end_node):
                edge_add = AlnEdge(self.nodes[node], self.end_node)
                edge_add.count = 1
                edge_add.score = 0
                self.add_edge(edge_add)  

        for node in node_to_remove:
            self.delete_node(node)
        

    def jsOutput(self):
        """returns a list of strings containing a a description of the graph for viz.js, http://visjs.org"""
        # get the backbone of the graph
        backbone_path = self.backbone_nodes
        pathdict = {}
        for i, nodeID in enumerate(list(backbone_path.values())):
            pathdict[nodeID] = i*150

        lines = ['var nodes = [']

        ni = self.nodes.values()
        count = 0
        for node in ni:
            line = '    {id:'+str(node.ID)+', label: "'+node.base+'"'
            if node.ID in pathdict and count % 5 == 0:
                line += ', allowedToMoveX: false, x: ' + str(pathdict[node.ID]) + ', y: 0 , allowedToMoveY: true },'
            else:
                line += '},'
            lines.append(line)

        lines[-1] = lines[-1][:-1]
        lines.append('];')

        lines.append(' ')

        lines.append('var edges = [')
        ni =  self.nodes.values()
        for node in ni:
            nodeID = str(node.ID)
            for edge in node._out_edges:
                #target = str(edge)
                #target =  "(%d) -> (%d) " % (edge.in_node.ID, edge.out_node.ID)
                target =  "%d" % (edge.out_node.ID)
                #weight = str(len(node._outEdges[edge].labels)+1)
                weight = str(edge.count)
                lines.append('    {from: '+nodeID+', to: '+target+", arrows:'to', value: "+weight+'},')
            '''for alignededge in node.alignedTo:
                # These edges indicate alignment to different bases, and are
                # undirected; thus make sure we only plot them once:
                if node.ID > alignededge:
                    continue
                target = str(alignededge)
                lines.append('    {from: '+nodeID+', to: '+target+', value: 1, style: "dash-line"},')'''
        lines[-1] = lines[-1][:-1]
        lines.append('];')
        return lines
    
    def htmlOutput(self, outfile):
        header = """
                  <!doctype html>
                  <html>
                  <head>
                    <title>POA Graph Alignment</title>
                    <style type="text/css">
                        html, body {
                        font: 10pt arial;
                        }
                        #mynetwork {
                        width: 1800px;
                        height: 600px;
                        border: 1px solid lightgray;
                        }
                    </style>
                    <script type="text/javascript" src="https://visjs.github.io/vis-network/standalone/umd/vis-network.min.js"></script>
                  </head>
                  <body>
                  <div id="mynetwork"></div>
                  <script type="text/javascript">
                    // create a network
                    var nodes = null;
                    var edges = null;
                    var network = null;
                    function draw() {
                  """
        outfile.write(textwrap.dedent(header[1:]))
        lines = self.jsOutput()
        for line in lines:
            outfile.write(line+'\n')
        footer = """
                      // Instantiate our network object.
                    var container = document.getElementById('mynetwork');
                    var data = {
                    nodes: nodes,
                    edges: edges
                    };
                    var options = {
                    nodes: {
                        shape: 'dot',
                        scaling: {
                        customScalingFunction: function (min,max,total,value) {
                            return value/total;
                        },
                        min:5,
                        max:150
                        }
                    }
                    };
                    network = new vis.Network(container, data, options);
                }
                </script>
                <body onload="draw()">
                </body>
                </html>
                """
        outfile.write(textwrap.dedent(footer))


import os

def store_labels_as_fa(labels, folder):
    

    os.makedirs(folder, exist_ok=True)
    for i, sequence in enumerate(labels):
        trimmed_sequence = sequence 
        if sequence[0] == 'E' or sequence[0] == 'B':
            trimmed_sequence = sequence[1:]

        if sequence[-1] == 'E' or sequence[-1] == 'B':
            trimmed_sequence = trimmed_sequence[:-1]

        trimmed_sequence_str = ''.join(trimmed_sequence)  
        reversed_sequence = trimmed_sequence_str[::-1]  
        
        graph_prefix = os.path.basename(os.path.dirname(folder))
        file_path = os.path.join(folder, 'output_{}_{}.fa'.format(graph_prefix, i))
        with open(file_path, 'w') as fa_file:
            fa_file.write('>{}_sequence{}    num:{} \n'.format(folder, i, len(trimmed_sequence)))
            fa_file.write(reversed_sequence + '\n')
    
    return file_path



def store_labels_as_fa_hanshuming(labels, folder, filename_prefix):
   
    os.makedirs(folder, exist_ok=True)

    for i, sequence in enumerate(labels):
        trimmed_sequence = sequence  

       
        if sequence[0] == 'E' or sequence[0] == 'B':
            trimmed_sequence = sequence[1:]
        if sequence[-1] == 'E' or sequence[-1] == 'B':
            trimmed_sequence = trimmed_sequence[:-1]

        trimmed_sequence_str = ''.join(trimmed_sequence) 
        reversed_sequence = trimmed_sequence_str[::-1]  

        
        file_path = os.path.join(folder, f'{filename_prefix}_{i}.fa')

        with open(file_path, 'w') as fa_file:
            fa_file.write(f'>{filename_prefix}_sequence{i}    num:{len(trimmed_sequence)}\n')
            fa_file.write(reversed_sequence + '\n')

    return file_path
